#include <ICM42670.h>

void ICM42670::Init(ICM42670_CONFIG icm42670_config){
	// ICM42670 Initialize
	icm42670_config.addr = icm42670_config.addr << 1;
	memcpy(&config, &icm42670_config, sizeof(ICM42670_CONFIG));

	// 센서 활성화
	uint8_t data[2] = {ICM42670_PWR_MGMT0, 0x9F};
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// 가속도 센서 설정
	data[0] = ICM42670_ACCEL_CONFIG0;
	data[1] = config.accel_range | config.accel_rate;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
	data[0] = ICM42670_ACCEL_CONFIG1;
	data[1] = config.accel_lpf;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// 자이로 센서 설정
	data[0] = ICM42670_GYRO_CONFIG0;
	data[1] = config.gyro_range | config.gyro_rate;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
	data[0] = ICM42670_GYRO_CONFIG1;
	data[1] = config.gyro_lpf;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
}

void ICM42670::Sleep(){

}
void ICM42670::Wakeup(){

}

void ICM42670::setAccelOffset(int16_t offset[]){
    // Read ICM42670_OFFSET_USER4
    uint8_t data[2] = {ICM42670_BLK_SEL_R, ICM42670_USER_BANK_MREG1};
    uint8_t offsetData[5] = {0};
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
    data[0] = ICM42670_MADDR_R;
    data[1] = ICM42670_OFFSET_USER4;
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
    data[0] = ICM42670_M_R;
    HAL_I2C_Master_Receive(&config.i2c, config.addr, &offsetData[0], 1, config.timeout);

    // Select Bank
    data[0] = ICM42670_BLK_SEL_W;
    data[1] = ICM42670_USER_BANK_MREG1;
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

    // Assemble offsetData
    offsetData[0] |= (offset[0] >> 4) & 0xF0;
    offsetData[1] = offset[0] & 0xFF;
    offsetData[2] = offset[1] & 0xFF;
    offsetData[3] = ((offset[1] >> 8) & 0x0F) | ((offset[2] >> 4) & 0xF0);
    offsetData[4] = offset[2] & 0xFF;

    // Write Offset Data
    for(uint8_t i=0;i<5;i++){
        data[0] = ICM42670_MADDR_W;
        data[1] = ICM42670_OFFSET_USER4 + i;
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
        data[0] = ICM42670_M_W;
        data[1] = offsetData[i];
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
    }
}
void ICM42670::setGyroOffset(int16_t offset[]){
    // Read ICM42670_OFFSET_USER4
    uint8_t data[2] = {ICM42670_BLK_SEL_R, ICM42670_USER_BANK_MREG1};
    uint8_t offsetData[5] = {0};
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
    data[0] = ICM42670_MADDR_R;
    data[1] = ICM42670_OFFSET_USER4;
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
    data[0] = ICM42670_M_R;
    HAL_I2C_Master_Receive(&config.i2c, config.addr, &offsetData[4], 1, config.timeout);

    // Select Bank
    data[0] = ICM42670_BLK_SEL_W;
    data[1] = ICM42670_USER_BANK_MREG1;
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

    // Assemble offsetData
    offsetData[0] = offset[0] & 0xFF;
    offsetData[1] = ((offset[0] >> 8) & 0x0F) | ((offset[1] >> 4) & 0xF0);
    offsetData[2] = offset[1] & 0xFF;
    offsetData[3] = offset[2] & 0xFF;
    offsetData[4] |= (offset[2] >> 8) & 0x0F;

    // Write Offset Data
    for(uint8_t i=0;i<5;i++){
        data[0] = ICM42670_MADDR_W;
        data[1] = ICM42670_OFFSET_USER0 + i;
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
        data[0] = ICM42670_M_W;
        data[1] = offsetData[i];
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
    }
}

int16_t *ICM42670::getRawAccel(){
	static int16_t output[3] = {0};
	uint8_t regAddr = ICM42670_ACCEL_DATA_X1, recv[6] = {0};
	uint16_t error = 0;
	while(HAL_I2C_IsDeviceReady(&config.i2c, config.addr, 10, config.timeout) != HAL_OK){SEGGER_RTT_printf(0, "getRawAccel\r\n");};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, recv, 6, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	while(HAL_I2C_GetState(&config.i2c) != HAL_I2C_STATE_READY){};
	for(uint8_t i=0;i<3;i++){
		output[i] = (recv[2*i] << 8) | recv[2*i + 1];
	}
	return output;
}

int16_t *ICM42670::getRawGyro(){
	static int16_t output[3] = {0};
	uint8_t regAddr = ICM42670_GYRO_DATA_X1, recv[6] = {0};
	uint16_t error = 0;
	while(HAL_I2C_IsDeviceReady(&config.i2c, config.addr, 10, config.timeout) != HAL_OK){SEGGER_RTT_printf(0, "getRawGyro\r\n");};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, recv, 6, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	while(HAL_I2C_GetState(&config.i2c) != HAL_I2C_STATE_READY){};
	for(uint8_t i=0;i<3;i++){
		output[i] = (recv[2*i] << 8) | recv[2*i + 1];
	}
	return output;
}
int16_t ICM42670::getRawTemp(){
	int16_t output = 0;
	uint8_t regAddr = ICM42670_TEMP_DATA1, recv[2] = {0};
	uint16_t error = 0;
	while(HAL_I2C_IsDeviceReady(&config.i2c, config.addr, 10, config.timeout) != HAL_OK){SEGGER_RTT_printf(0, "getRawTemp\r\n");};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, recv, 2, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	while(HAL_I2C_GetState(&config.i2c) != HAL_I2C_STATE_READY){};
	output = recv[0] << 8 | recv[1];
	return output;
}

uint8_t ICM42670::getDeviceId(){
	// 디바이스 ID 리턴
	uint8_t regAddr = ICM42670_WHO_AM_I, deviceId = 0;
	uint16_t error = 0;
	error = HAL_I2C_GetState(&config.i2c);
	while(error != HAL_I2C_STATE_READY){};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, &deviceId, 1, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	return deviceId;
}

int16_t *ICM42670::getAccelOffset(){
    static int16_t output[3] = {0};

    // Select USERBANK1 and Address
    uint8_t data[2] = {ICM42670_BLK_SEL_R, ICM42670_USER_BANK_MREG1};
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

    // MREG Read Operation
    uint8_t recvData[5] = {0}, regAddr = ICM42670_M_R;
    for(uint8_t i=0;i<5;i++){
        data[0] = ICM42670_MADDR_R;
        data[1] = ICM42670_OFFSET_USER4 + i;
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, &regAddr, 1, config.timeout);
        HAL_I2C_Master_Receive(&config.i2c, config.addr, &recvData[i], 1, config.timeout);
    }

    // Assemble data
    output[0] = ((recvData[0] & 0xF0) << 4) | recvData[1];
    output[1] = ((recvData[3] & 0x0F) << 8) | recvData[2];
    output[2] = ((recvData[3] & 0xF0) << 4) | recvData[4];
    return output;
}

int16_t *ICM42670::getGyroOffset(){
    static int16_t output[3] = {0};

    // Select USERBANK1 and Address
    uint8_t data[2] = {ICM42670_BLK_SEL_R, ICM42670_USER_BANK_MREG1};
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

    // MREG Read Operation
    uint8_t recvData[5] = {0}, regAddr = ICM42670_M_R;
    for(uint8_t i=0;i<5;i++){
        data[0] = ICM42670_MADDR_R;
        data[1] = ICM42670_OFFSET_USER0 + i;
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
        HAL_I2C_Master_Transmit(&config.i2c, config.addr, &regAddr, 1, config.timeout);
        HAL_I2C_Master_Receive(&config.i2c, config.addr, &recvData[i], 1, config.timeout);
    }

    // Assemble data
    output[0] = ((recvData[1] & 0x0F) << 8) | recvData[0];
    output[1] = ((recvData[1] & 0xF0) << 4) | recvData[2];
    output[2] = ((recvData[4] & 0x0F) << 8) | recvData[3];
    return output;
}

int16_t *ICM42670::computeAccelOffset(uint16_t samples){
    uint16_t counter = 0;
    static int16_t output[3] = {0};
    memset(output, 0x00, sizeof(output));
    do{
        int16_t *accel = getRawAccel();
        for(uint8_t i=0;i<3;i++){
            if(counter > 0){
                output[i] = (((float)counter-1.0f)/(float)counter)*(float)output[i] + (1.0f/(float)counter)*(float)accel[i];
            }else{
                output[i] = accel[i];
            }
        }
    }while(++counter<samples);
    return output;
}

int16_t *ICM42670::computeGyroOffset(uint16_t samples){
    uint16_t counter = 0;
    static int16_t output[3] = {0};
    memset(output, 0x00, sizeof(output));
    do{
        int16_t *gyro = getRawGyro();
        for(uint8_t i=0;i<3;i++){
            if(counter > 0){
                output[i] = (((float)counter-1.0f)/(float)counter)*(float)output[i] + (1.0f/(float)counter)*(float)gyro[i];
            }else{
                output[i] = gyro[i];
            }
        }
    }while(++counter<samples);
    return output;
}
