/*
        ICM42670 Library
        Author : Seonguk Jeong
        Created at : 2023-10-1
        Updated at : 2023-10-1
        Target : STM32
*/

#ifndef ROVERDYN_STM32_LIBRARY_SENSORS_ICM42670_H_
#define ROVERDYN_STM32_LIBRARY_SENSORS_ICM42670_H_

#include <stdio.h>
#include <stdint.h>
#include <String.h>
#include <Configure.h>

// Register definitions
#define ICM42670_MCLK_RDY                0x00
#define ICM42670_DEVICE_CONFIG           0x01
#define ICM42670_SIGNAL_PATH_REST        0x02
#define ICM42670_DRIVE_CONFIG1           0x03
#define ICM42670_DRIVE_CONFIG2           0x04
#define ICM42670_DRIVE_CONFIG3           0x05
#define ICM42670_INT_CONFIG              0x06
#define ICM42670_TEMP_DATA1              0x09
#define ICM42670_TEMP_DATA0              0x0A
#define ICM42670_ACCEL_DATA_X1           0x0B
#define ICM42670_ACCEL_DATA_X0           0x0C
#define ICM42670_ACCEL_DATA_Y1           0x0D
#define ICM42670_ACCEL_DATA_Y0           0x0E
#define ICM42670_ACCEL_DATA_Z1           0x0F
#define ICM42670_ACCEL_DATA_Z0           0x10
#define ICM42670_GYRO_DATA_X1            0x11
#define ICM42670_GYRO_DATA_X0            0x12
#define ICM42670_GYRO_DATA_Y1            0x13
#define ICM42670_GYRO_DATA_Y0            0x14
#define ICM42670_GYRO_DATA_Z1            0x15
#define ICM42670_GYRO_DATA_Z0            0x16
#define ICM42670_TMST_FSYNCH             0x17
#define ICM42670_TMST_FSYNCL             0x18
#define ICM42670_APEX_DATA4              0x1D
#define ICM42670_APEX_DATA5              0x1E
#define ICM42670_PWR_MGMT0               0x1F
#define ICM42670_GYRO_CONFIG0            0x20
#define ICM42670_ACCEL_CONFIG0           0x21
#define ICM42670_TEMP_CONFIG0            0x22
#define ICM42670_GYRO_CONFIG1            0x23
#define ICM42670_ACCEL_CONFIG1           0x24
#define ICM42670_APEX_CONFIG0            0x25
#define ICM42670_APEX_CONFIG1            0x26
#define ICM42670_WOM_CONFIG              0x27
#define ICM42670_FIFO_CONFIG1            0x28
#define ICM42670_FIFO_CONFIG2            0x29
#define ICM42670_FIFO_CONFIG3            0x30
#define ICM42670_INT_SOURCE0             0x2B
#define ICM42670_INT_SOURCE1             0x2C
#define ICM42670_INT_SOURCE3             0x2D
#define ICM42670_INT_SOURCE4             0x2E
#define ICM42670_FIFO_LOST_PKT0          0x2F
#define ICM42670_FIFO_LOST_PKT1          0x30
#define ICM42670_APEX_DATA0              0x31
#define ICM42670_APEX_DATA1              0x32
#define ICM42670_APEX_DATA2              0x33
#define ICM42670_APEX_DATA3              0x34
#define ICM42670_INTF_CONFIG0            0x35
#define ICM42670_INTF_CONFIG1            0x36
#define ICM42670_INT_STATUS_DRDY         0x39
#define ICM42670_INT_STATUS              0x3A
#define ICM42670_INT_STATUS2             0x3B
#define ICM42670_INT_STATUS3             0x3C
#define ICM42670_FIFO_COUNTH             0x3D
#define ICM42670_FIFO_COUNTL             0x3E
#define ICM42670_FIFO_DATA               0x3F
#define ICM42670_WHO_AM_I                0x75
#define ICM42670_BLK_SEL_W               0x79
#define ICM42670_MADDR_W                 0x7A
#define ICM42670_M_W                     0x7B
#define ICM42670_BLK_SEL_R               0x7C
#define ICM42670_MADDR_R                 0x7D
#define ICM42670_M_R                     0x7E

// USER BANK 1
#define ICM42670_TMST_CONFIG1            0x00
#define ICM42670_FIFO_CONFIG5            0x01
#define ICM42670_FIFO_CONFIG6            0x02
#define ICM42670_FSYNC_CONFIG            0x03
#define ICM42670_INT_CONFIG0             0x04
#define ICM42670_INT_CONFIG1             0x05
#define ICM42670_SENSOR_CONFIG3          0x06
#define ICM42670_ST_CONFIG               0x13
#define ICM42670_SELFTEST                0x14
#define ICM42670_INTF_CONFIG6            0x23
#define ICM42670_INTF_CONFIG10           0x25
#define ICM42670_INTF_CONFIG7            0x28
#define ICM42670_OTP_CONFIG              0x2B
#define ICM42670_INT_SOURCE6             0x2F
#define ICM42670_INT_SOURCE7             0x30
#define ICM42670_INT_SOURCE8             0x31
#define ICM42670_INT_SOURCE9             0x32
#define ICM42670_INT_SOURCE10            0x33
#define ICM42670_APEX_CONFIG2            0x44
#define ICM42670_APEX_CONFIG3            0x45
#define ICM42670_APEX_CONFIG4            0x46
#define ICM42670_APEX_CONFIG5            0x47
#define ICM42670_APEX_CONFIG9            0x48
#define ICM42670_APEX_CONFIG10           0x49
#define ICM42670_APEX_CONFIG11           0x4A
#define ICM42670_ACCEL_WOM_X_THR         0x4B
#define ICM42670_ACCEL_WOM_Y_THR         0x4C
#define ICM42670_ACCEL_WOM_Z_THR         0x4D
#define ICM42670_OFFSET_USER0            0x4E
#define ICM42670_OFFSET_USER1            0x4F
#define ICM42670_OFFSET_USER2            0x50
#define ICM42670_OFFSET_USER3            0x51
#define ICM42670_OFFSET_USER4            0x52
#define ICM42670_OFFSET_USER5            0x53
#define ICM42670_OFFSET_USER6            0x54
#define ICM42670_OFFSET_USER7            0x55
#define ICM42670_OFFSET_USER8            0x56
#define ICM42670_ST_STATUS1              0x63
#define ICM42670_ST_STATUS2              0x64
#define ICM42670_FDR_CONFIG              0x66
#define ICM42670_APEX_CONFIG12           0x67

// USER BANK 2
#define ICM42670_OPT_CTRL7               0x06

// USER BANK3
#define ICM42670_XA_ST_DATA              0x00
#define ICM42670_YA_ST_DATA              0x01
#define ICM42670_ZA_ST_DATA              0x02
#define ICM42670_XG_ST_DATA              0x03
#define ICM42670_YG_ST_DATA              0x04
#define ICM42670_ZG_ST_DATA              0x05

// Pre-defined settings
#define ICM42670_USER_BANK_MREG1        0x00
#define ICM42670_USER_BANK_MREG2        0x28
#define ICM42670_USER_BANK_MREG3        0x50
#define ICM42670_GYRO_RANGE_250		(0b11 << 5)
#define ICM42670_GYRO_RANGE_500		(0b10 << 5)
#define ICM42670_GYRO_RANGE_1000	(0b01 << 5)
#define ICM42670_GYRO_RANGE_2000	(0b00 << 5)
#define ICM42670_GYRO_ODR_12		0b1100
#define ICM42670_GYRO_ODR_25		0b1011
#define ICM42670_GYRO_ODR_50		0b1010
#define ICM42670_GYRO_ODR_100		0b1001
#define ICM42670_GYRO_ODR_200		0b1000
#define ICM42670_GYRO_ODR_400		0b0111
#define ICM42670_GYRO_ODR_800		0b0110
#define ICM42670_GYRO_ODR_1600		0b0101
#define ICM42670_GYRO_LPF_16HZ		0b0111
#define ICM42670_GYRO_LPF_25HZ		0b0110
#define ICM42670_GYRO_LPF_34HZ		0b0101
#define ICM42670_GYRO_LPF_53HZ		0b0100
#define ICM42670_GYRO_LPF_73HZ		0b0011
#define ICM42670_GYRO_LPF_121HZ		0b0010
#define ICM42670_GYRO_LPF_180HZ		0b0001
#define ICM42670_GYRO_LPF_DISABLE	0b0000
#define ICM42670_ACCEL_RANGE_2		(0b11 << 5)
#define ICM42670_ACCEL_RANGE_4		(0b10 << 5)
#define ICM42670_ACCEL_RANGE_8		(0b01 << 5)
#define ICM42670_ACCEL_RANGE_16		(0b00 << 5)
#define ICM42670_ACCEL_ODR_12		0b1100
#define ICM42670_ACCEL_ODR_25		0b1011
#define ICM42670_ACCEL_ODR_50		0b1010
#define ICM42670_ACCEL_ODR_100		0b1001
#define ICM42670_ACCEL_ODR_200		0b1000
#define ICM42670_ACCEL_ODR_400		0b0111
#define ICM42670_ACCEL_ODR_800		0b0110
#define ICM42670_ACCEL_ODR_1600		0b0101
#define ICM42670_ACCEL_LPF_16HZ		0b0111
#define ICM42670_ACCEL_LPF_25HZ		0b0110
#define ICM42670_ACCEL_LPF_34HZ		0b0101
#define ICM42670_ACCEL_LPF_53HZ		0b0100
#define ICM42670_ACCEL_LPF_73HZ		0b0011
#define ICM42670_ACCEL_LPF_121HZ	0b0010
#define ICM42670_ACCEL_LPF_180HZ	0b0001
#define ICM42670_ACCEL_LPF_DISABLE	0b0000

// Address definitions
#define ICM42670_ADDR0  0x68
#define ICM42670_ADDR1  0x69

struct ICM42670_CONFIG{
		I2C_HandleTypeDef i2c;
        uint8_t addr;
        uint32_t timeout = 10000;
        uint8_t accel_range;
        uint8_t accel_rate;
        uint8_t accel_lpf;
        uint8_t gyro_range;
        uint8_t gyro_rate;
        uint8_t gyro_lpf;
};
class ICM42670{
	private:
		ICM42670_CONFIG config;
	public:
		void Init(ICM42670_CONFIG icm42670_config);
		void Sleep();
		void Wakeup();
		void setAccelOffset(int16_t offset[]);
		void setGyroOffset(int16_t offset[]);
		int16_t *getRawAccel();
		int16_t *getRawGyro();
		int16_t getRawTemp();
		uint8_t getDeviceId();
		int16_t *getAccelOffset();
		int16_t *getGyroOffset();
		int16_t *computeAccelOffset(uint16_t samples);
		int16_t *computeGyroOffset(uint16_t samples);
};
#endif /* ROVERDYN_STM32_LIBRARY_SENSORS_ICM42670_H_ */
